"""
test_charex
~~~~~~~~~~~
"""
import json

import pytest

from charex import charex as c


# Test Character.
def test_character_init():
    """Given a string containing one or more codepoints representing
    a character, a :class:`Character` object is initialized.
    """
    exp_value = 'a'
    act = c.Character(exp_value)
    assert act.value == exp_value


def test_character_category():
    """When called, :attr:`Character.category` returns the Unicode
    category for the character.
    """
    char = c.Character('a')
    assert char.category == 'Lowercase Letter'


def test_character_code_point():
    """When called, :attr:`Character.code_point` returns the Unicode
    code point for the character.
    """
    char = c.Character('<')
    assert char.code_point == 'U+003C'


def test_character_decimal():
    """When called, :attr:`Character.decimal` gives the numeric value of
    the character if it has one. If the character does not have a
    numeric value, return None.
    """
    char = c.Character('2')
    assert char.decimal == 2

    char = c.Character('a')
    assert char.decimal is None


def test_character_decomposition():
    """When called, :attr:`Character.decomposition` returns the Unicode
    decomposition for the character.
    """
    char = c.Character('å')
    assert char.decomposition == '0061 030A'


def test_character_digit():
    """When called, :attr:`Character.digit` gives the numeric value of
    the character if it has one. If the character does not have a
    numeric value, return None.
    """
    char = c.Character('2')
    assert char.digit == 2

    char = c.Character('a')
    assert char.digit is None


def test_character_encode():
    """When called with a valid character encoding,
    :meth:`Character.is_normal` returns a hexadecimal string
    of the encoded form of the character.
    """
    char = c.Character('å')
    assert char.encode('utf8') == 'C3A5'


def test_character_escape_url():
    """When called with a valid character escaping scheme,
    :meth:`Character.escape` returns a string of the escaped
    form of the character.
    """
    # Percent encoding for URLs.
    char = c.Character('å')
    assert char.escape('url', 'utf8') == '%C3%A5'


def test_character_escape_html():
    """When called with a valid character escaping scheme,
    :meth:`Character.escape` returns a string of the escaped
    form of the character.
    """
    # Percent encoding for URLs.
    char = c.Character('å')
    assert char.escape('html') == '&aring;'


def test_character_is_normal():
    """When called with a valid normalization form,
    :meth:`Character.is_normal` returns whether the value
    is normalized for that form.
    """
    char = c.Character('a')
    assert char.is_normal('NFC')

    char = c.Character('å')
    assert not char.is_normal('NFD')


def test_character_name():
    """When called, :attr:`Character.name` returns the Unicode name
    for the code point.
    """
    char = c.Character('a')
    assert char.name == 'LATIN SMALL LETTER A'


def test_character_name_null():
    """When called, :attr:`Character.name` returns the Unicode name
    for the code point.
    """
    char = c.Character('\u0000')
    assert char.name == '<NULL>'


def test_character_name_private_use():
    """When called, :attr:`Character.name` returns the Unicode name
    for the code point.
    """
    char = c.Character('\ue90a')
    assert char.name == 'PRIVATE USE CHARACTER'


def test_character_numeric():
    """When called, :attr:`Character.numeric` gives the numeric value of
    the character if it has one. If the character does not have a
    numeric value, return None.
    """
    char = c.Character('2')
    assert char.numeric == 2

    char = c.Character('a')
    assert char.numeric is None


def test_character_normalize():
    """When given a normalization form, :meth:`Character.normalize` should
    return the normalized form of the character.
    """
    char = c.Character('å')
    assert char.normalize('NFD') == b'a\xcc\x8a'.decode('utf8')


def test_character_repr():
    """When called, :meth:`Character.__repr__` returns the Unicode code
    point and name for the code point.
    """
    char = c.Character('a')
    assert repr(char) == 'U+0061 (LATIN SMALL LETTER A)'


def test_character_denormalize():
    """When given a normalization form, :meth:`Character.reverse_normalize`
    should return the normalized form of the character.
    """
    exp = ("\uf907", "\uf908", "\uface")
    char = c.Character('\u9f9c')
    assert char.denormalize('nfc') == exp


def test_character_summarize():
    """When called, :meth:`Character.summarize` returns a summary of the
    character's information as a :class:`str`.
    """
    exp = 'a U+0061 (LATIN SMALL LETTER A)'
    char = c.Character('a')
    assert char.summarize() == exp


def test_character_summarize_control():
    """When called, :meth:`Character.summarize` returns a summary of the
    character's information as a :class:`str`.
    """
    exp = '\u240a U+000A (<LINE FEED (LF)>)'
    char = c.Character('\n')
    assert char.summarize() == exp


# Test Lookup.
def test_lookup_init_set_source():
    """Given a key for a data file, an instance of Lookup should be
    created with the data file loaded.
    """
    exp_source = 'rev_nfc'
    with open(f'charex/data/{exp_source}.json') as fh:
        data = json.load(fh)
        exp_data = {k: tuple(data[k]) for k in data}
    act = c.Lookup(exp_source)
    assert act.source == exp_source
    assert act.data == exp_data


def test_lookup_query():
    """Given a string, :meth:`Lookup.query` should return the value
    for that string from the loaded data.
    """
    exp = ("\uf907", "\uf908", "\uface")
    key = '\u9f9c'
    lkp = c.Lookup('rev_nfc')
    act = lkp.query(key)
    assert act == exp

    # If key is not present in the data, return an empty tuple.
    key = 'a'
    assert lkp.query(key) == ()



"""
normal
~~~~~~

Functions for normalizing strings.
"""
from collections.abc import Callable
from itertools import permutations
from json import dumps
import unicodedata as ucd

from charex import util


# Constants.
LEN_UNICODE = 0x10FFFF


# Registry.
forms = {}


# Registration.
class reg_form:
    """A decorator for registering normalization forms.

    :param key: The name the normalization form is registered under.
    """
    def __init__(self, key: str) -> None:
        self.key = key

    def __call__(
        self,
        fn: Callable[[str], str]
    ) -> Callable[[str], str]:
        forms[self.key] = fn
        return fn


# Utility functions.
def build_denormalization_map(norm_fn: Callable[[str], str]) -> str:
    """Create a JSON string mapping each Unicode character to the
    other Unicode characters that normalize to it.

    :param norm_fn: The normalization function.
    :return: The denormalization map as a JSON :class:`str`.
    :rtype: str
    """
    # The denormalization map.
    dn_map: dict[str, set[str]] = {}

    # Process every Unicode character.
    for n in range(LEN_UNICODE):
        base = chr(n)

        # If the character normalizes to a different character,
        # add that relationship to the map.
        normal = norm_fn(base)
        if normal and normal != base:
            dn_map.setdefault(normal, set())
            dn_map[normal].add(base)

        # If the character decomposes, add the relationship between
        # the character and its possible decompositions to the map.
        decomp = form_nfd(base)
        if len(decomp) > 1:
            root, marks = decomp[0], decomp[1:]
            for mut in permutations(marks):
                rmut = root + ''.join(mut)
                normal = norm_fn(rmut)
                if normal != rmut:
                    dn_map.setdefault(normal, set())
                    dn_map[normal].add(rmut)

    # Test to ensure decompositions are accurate and not redundant.
    for normal in dn_map:
        for base in dn_map[normal]:
            actual = norm_fn(base)
            assert actual == normal
            assert actual != base
            assert normal != base

    # Return the map as a JSON string.
    listed = {k: list(dn_map[k]) for k in dn_map}
    ordered = {k: sorted(listed[k]) for k in listed}
    return dumps(ordered, indent=4)


def find_max_decomposition() -> tuple[str, int]:
    """Find the longest single character decomposition.

    :return: The character and the number of characters in the
        decomposition as a :class:`tuple`.
    :rtype: tuple
    """
    long_char = ''
    decomp_max = 0
    for n in range(LEN_UNICODE):
        char = chr(n)
        decomp_char = form_nfd(char)
        decomp_len = len(decomp_char)
        if decomp_len > decomp_max:
            long_char = char
            decomp_max = decomp_len
    return long_char, decomp_max


def get_description(formkey: str) -> str:
    """Get the description for the normalization form.

    :param formkey: The key for the form in the form registry.
    :return: The description as a :class:`str`.
    :rtype: str
    """
    form = forms[formkey]
    return util.get_description_from_docstring(form)


def get_forms() -> tuple[str, ...]:
    """Return the keys of the registered normalization forms.

    :return: The names of the normalization forms as a :class:`tuple`.
    :rtype: tuple
    """
    return tuple(form for form in forms)


# Normalization function.
def normalize(formkey: str, base: str) -> str:
    """Normalize the base string with the form.

    :param formkey: The key of a registered normalization form.
    :param base: The string to normalize.
    :return: The normalized :class:`str`.
    :rtype: str
    """
    form = forms[formkey]
    return form(base)


# Normalization forms.
@reg_form('casefold')
def form_casefold(base: str) -> str:
    """Remove all case distinctions from the string.

    :param base: The string to normalize.
    :return: The normalized :class:`str`.
    :rtype: str
    """
    return str.casefold(base)


@reg_form('nfc')
def form_nfc(base: str) -> str:
    """Normalization form composition.

    :param base: The string to normalize.
    :return: The normalized :class:`str`.
    :rtype: str
    """
    return ucd.normalize('NFC', base)


@reg_form('nfd')
def form_nfd(base: str) -> str:
    """Normalization form decomposition.

    :param base: The string to normalize.
    :return: The normalized :class:`str`.
    :rtype: str
    """
    return ucd.normalize('NFD', base)


@reg_form('nfkc')
def form_nfkc(base: str) -> str:
    """Normalization form compatibility composition.

    :param base: The string to normalize.
    :return: The normalized :class:`str`.
    :rtype: str
    """
    return ucd.normalize('NFKC', base)


@reg_form('nfkd')
def form_nfkd(base: str) -> str:
    """Normalization form compatibility decomposition.

    :param base: The string to normalize.
    :return: The normalized :class:`str`.
    :rtype: str
    """
    return ucd.normalize('NFKD', base)



"""
shell
~~~~~

An interactive command shell for :mod:`charex`.
"""
from collections.abc import Callable, Sequence
from argparse import (
    ArgumentParser,
    Namespace,
    _SubParsersAction,
    RawDescriptionHelpFormatter
)
from cmd import Cmd
from itertools import zip_longest
import readline
from shlex import split
from shutil import get_terminal_size
from textwrap import wrap

from charex import charex as ch
from charex import charsets as cset
from charex import denormal as dn
from charex import escape as esc
from charex import normal as nl
from charex import util


# Utility functions.
def make_description_row(name: str, namewidth: int, descr: str) -> str:
    """Create a two column row with a name and description.

    :param name: The content for the first column.
    :param namewidth: The width of the first column.
    :param descr: The content for the second column.
    :return: The row as a :class:`str`.
    :rtype: str
    """
    name_lines = wrap(name, namewidth)
    descr_lines = wrap(descr, 77 - namewidth)
    lines = (
        f'{n:<{namewidth}}  {d}'
        for n, d in zip_longest(name_lines, descr_lines, fillvalue='')
    )
    return '\n'.join(lines)


def write_list(
    items: Sequence[str],
    get_descr: Callable[[str], str],
    show_descr: bool
) -> None:
    """Output the given list.

    :param items: The items to list.
    :param get_descr: The function used to look up the discription
        of each item.
    :param show_descr: Whether include the descriptions of each item
        in the output.
    :return: None.
    :rtype: NoneType
    """
    width = max(len(item) for item in items)
    for item in items:
        if show_descr:
            descr = get_descr(item)
            row = make_description_row(item, width, descr)
            print(row)
            print()
        else:
            print(item)


# Running modes.
def mode_cd(args: Namespace) -> None:
    """Decode the given address in all codecs.

    :param args: The arguments used when the script was invoked.
    :return: None.
    :rtype: NoneType
    """
    # Normalize the data.
    if args.base.startswith('0b'):
        b = util.bin2bytes(args.base[2:])
    elif args.base.startswith('0x'):
        b = util.hex2bytes(args.base[2:])
    else:
        b = args.base.encode('utf8')

    # Get the data.
    codecs = cset.get_codecs()
    results = cset.multidecode(b, (codec for codec in codecs))

    # Write the output.
    width = max(len(codec) for codec in codecs)
    for key in results:
        c = results[key]
        details = ''
        if len(c) < 1:
            details = '*** no character ***'
        elif len(c) > 1:
            details = '*** multiple characters ***'
        else:
            char = ch.Character(c)
            details = f'{char.code_point} {char.name}'
        c = util.neutralize_control_characters(c)
        print(f'{key:>{width}}: {c} {details}')
    print()


def mode_ce(args: Namespace) -> None:
    """Encode the given character in all codecs.

    :param args: The arguments used when the script was invoked.
    :return: None.
    :rtype: NoneType
    """
    # Get the data.
    codecs = cset.get_codecs()
    results = cset.multiencode(args.base, (codec for codec in codecs))

    # Write the output.
    width = max(len(codec) for codec in codecs)
    for key in results:
        if b := results[key]:
            c = ''.join(f'{n:>02x}'.upper() for n in b)
            print(f'{key:>{width}}: {c}')
    print()


def mode_cl(args: Namespace) -> None:
    """List registered character sets.

    :param args: The arguments used when the script was invoked.
    :return: None.
    :rtype: NoneType
    """
    codecs = cset.get_codecs()
    write_list(codecs, cset.get_codec_description, args.description)
    print()


def mode_ct(args: Namespace) -> None:
    """Count denormalization results.

    :param args: The arguments used when the script was invoked.
    :return: None.
    :rtype: NoneType
    """
    count = dn.count_denormalizations(args.base, args.form, args.maxdepth)
    print(f'{count:,}')
    print()


def mode_dm(args: Namespace) -> None:
    """Build a denormalization map.

    :param args: The arguments used when the script was invoked.
    :return: None.
    :rtype: NoneType
    """
    normal_fn = nl.forms[args.form]
    results = nl.build_denormalization_map(normal_fn)
    print(results)
    print()


def mode_dn(args: Namespace) -> None:
    """Perform denormalizations.

    :param args: The arguments used when the script was invoked.
    :return: None.
    :rtype: NoneType
    """
    results = dn.denormalize(
        args.base,
        args.form,
        args.maxdepth,
        args.number,
        args.random,
        args.seed
    )
    for result in results:
        print(result)
    print()


def mode_dt(args: Namespace) -> None:
    """Display details for a code point.

    :param args: The arguments used when the script was invoked.
    :return: None.
    :rtype: NoneType
    """
    def rev_normalize(char: ch.Character, form: str) -> str:
        points = char.denormalize(form)
        values = []
        for point in points:
            if len(point) == 1:
                c = ch.Character(point)
                values.append(c.summarize())
            elif len(point) > 1:
                values.append(f'{point} *** multiple characters ***')
                for item in point:
                    char = ch.Character(item)
                    values.append('  ' + char.summarize())
        if not values:
            return ''
        return ('\n' + ' ' * 22).join(v for v in values)

    # Gather the details for display.
    char = ch.Character(args.codepoint)
    details = (
        ('Display', char.value),
        ('Name', char.name),
        ('Code Point', char.code_point),
        ('Category', char.category),
        ('UTF-8', char.encode('utf8')),
        ('UTF-16', char.encode('utf_16_be')),
        ('UTF-32', char.encode('utf_32_be')),
        ('Decomposition', char.decomposition),
        ('C encoded', char.escape('c')),
        ('URL encoded', char.escape('url')),
        ('HTML encoded', char.escape('html')),
        ('Reverse Cfold', rev_normalize(char, 'casefold')),
        ('Reverse NFC', rev_normalize(char, 'nfc')),
        ('Reverse NFD', rev_normalize(char, 'nfd')),
        ('Reverse NFKC', rev_normalize(char, 'nfkc')),
        ('Reverse NFKD', rev_normalize(char, 'nfkd')),
    )

    # Display the details.
    width = 20
    for detail in details:
        label, value = detail
        if value:
            print(f'{label:>{width}}: {value}')
    print()


def mode_el(args: Namespace) -> None:
    """List the registered escape schemes.

    :param args: The arguments used when the script was invoked.
    :return: None.
    :rtype: NoneType
    """
    schemes = esc.get_schemes()
    write_list(schemes, esc.get_description, args.description)
    print()


def mode_es(args: Namespace) -> None:
    """Escape a string using the given scheme.

    :param args: The arguments used when the script was invoked.
    :return: None.
    :rtype: NoneType
    """
    result = esc.escape(args.base, args.scheme, 'utf8')
    print(result)
    print()


def mode_fl(args: Namespace) -> None:
    """List registered normalization forms.

    :param args: The arguments used when the script was invoked.
    :return: None.
    :rtype: NoneType
    """
    forms = nl.get_forms()
    write_list(forms, nl.get_description, args.description)
    print()


def mode_nl(args: Namespace) -> None:
    """Perform normalizations.

    :param args: The arguments used when the script was invoked.
    :return: None.
    :rtype: NoneType
    """
    result = nl.normalize(args.form, args.base)
    print(result)
    print()


def mode_sh(args: Namespace | None) -> None:
    """Run in an interactive shell.

    :param args: The arguments used when the script was invoked.
    :return: None.
    :rtype: NoneType
    """
    Shell(completekey='tab').cmdloop()


# Command parsing.
def build_parser() -> ArgumentParser:
    """Build the argument parser.

    :return: The :class:`argparse.ArgumentParser`.
    :rtype: argparse.ArgumentParser
    """
    # Build the argument parser.
    p = ArgumentParser(
        description='Unicode and character set explorer.',
        epilog=describe_modes(),
        formatter_class=RawDescriptionHelpFormatter,
        prog='charex'
    )

    # Build subparsers for each mode.
    spa = p.add_subparsers(
        help=list_modes(),
        metavar='mode',
        required=True
    )
    parse_cd(spa)
    parse_ce(spa)
    parse_cl(spa)
    parse_ct(spa)
    parse_dm(spa)
    parse_dn(spa)
    parse_dt(spa)
    parse_el(spa)
    parse_es(spa)
    parse_fl(spa)
    parse_nl(spa)
    parse_sh(spa)

    return p


def parse_cd(spa: _SubParsersAction) -> None:
    """Add the cd mode subparser.

    :param spa: The subparser action used to add a new subparser to
        the main parser.
    :return: None.
    :rtype: NoneType
    """
    sp = spa.add_parser(
        'cd',
        description='Decode the given address in all codecs.'
    )
    sp.add_argument(
        'base',
        help=(
            'The base integer. Prefix the integer with "0x" for hex '
            'or "0b" for binary. No prefix will be interpreted as the'
            'UTF-8 address of the character.'
        ),
        action='store',
        type=str
    )
    sp.set_defaults(func=mode_cd)


def parse_ce(spa: _SubParsersAction) -> None:
    """Add the ce mode subparser.

    :param spa: The subparser action used to add a new subparser to
        the main parser.
    :return: None.
    :rtype: NoneType
    """
    sp = spa.add_parser(
        'ce',
        description='Encode the given character in all codecs.'
    )
    sp.add_argument(
        'base',
        help='The character to lookup in each character set.',
        action='store',
        type=str
    )
    sp.set_defaults(func=mode_ce)


def parse_cl(spa: _SubParsersAction) -> None:
    """Add the charsetlist mode subparser.

    :param spa: The subparser action used to add a new subparser to
        the main parser.
    :return: None.
    :rtype: NoneType
    """
    sp = spa.add_parser(
        'cl',
        aliases=['charsetlist', 'csetlist', 'cslist'],
        description='List the registered character sets.'
    )
    sp.add_argument(
        '-d', '--description',
        help='Show the description for the character sets.',
        action='store_true'
    )
    sp.set_defaults(func=mode_cl)


def parse_ct(spa: _SubParsersAction) -> None:
    """Add the ct mode subparser.

    :param spa: The subparser action used to add a new subparser to
        the main parser.
    :return: None.
    :rtype: NoneType
    """
    sp = spa.add_parser(
        'ct',
        aliases=['count',],
        description='Count of denormalization results.'
    )
    sp.add_argument(
        'form',
        help='The Unicode normalization form for the denormalization.',
        choices=('nfc', 'nfd', 'nfkc', 'nfkd',)
    )
    sp.add_argument(
        'base',
        help='The base normalized string.',
        action='store',
        type=str
    )
    sp.add_argument(
        '-m', '--maxdepth',
        help=(
            'Maximum number of reverse normalizations to use '
            'for each character.'
        ),
        default=0,
        action='store',
        type=int
    )
    sp.set_defaults(func=mode_ct)


def parse_dm(spa: _SubParsersAction) -> None:
    """Add the dm mode subparser.

    :param spa: The subparser action used to add a new subparser to
        the main parser.
    :return: None.
    :rtype: NoneType
    """
    valid_forms = ', '.join(form for form in nl.get_forms())

    sp = spa.add_parser(
        'dm',
        aliases=['denormalmap',],
        description='Build a denormalization map.'
    )
    sp.add_argument(
        'form',
        choices=nl.get_forms(),
        help=(
            'The normalization form for the normalization. Valid '
            f'options are: {valid_forms}.'
        ),
        metavar='form'
    )
    sp.set_defaults(func=mode_dm)


def parse_dn(spa: _SubParsersAction) -> None:
    """Add the dn mode subparser.

    :param spa: The subparser action used to add a new subparser to
        the main parser.
    :return: None.
    :rtype: NoneType
    """
    valid_forms = ', '.join(form for form in nl.get_forms())

    sp = spa.add_parser(
        'dn',
        aliases=['denormal',],
        description='Denormalize a string.'
    )
    sp.add_argument(
        'form',
        choices=nl.get_forms(),
        help=(
            'The normalization form for the denormalization. Valid '
            f'options are: {valid_forms}.'
        ),
        metavar='form'
    )
    sp.add_argument(
        'base',
        help='The base normalized string.',
        action='store',
        type=str
    )
    sp.add_argument(
        '-m', '--maxdepth',
        help=(
            'Maximum number of reverse normalizations to use '
            'for each character.'
        ),
        default=0,
        action='store',
        type=int
    )
    sp.add_argument(
        '-n', '--number',
        help='Maximum number of results to return.',
        default=0,
        action='store',
        type=int
    )
    sp.add_argument(
        '-r', '--random',
        help='Randomize the denormalization.',
        action='store_true'
    )
    sp.add_argument(
        '-s', '--seed',
        help='Seed the randomized denormalization.',
        action='store',
        default=''
    )
    sp.set_defaults(func=mode_dn)


def parse_dt(spa: _SubParsersAction) -> None:
    """Add the dt mode subparser.

    :param spa: The subparser action used to add a new subparser to
        the main parser.
    :return: None.
    :rtype: NoneType
    """
    sp = spa.add_parser(
        'dt',
        aliases=['details',],
        description='Display the details for the given code point.'
    )
    sp.add_argument(
        'codepoint',
        help='The code point.',
        action='store',
        type=str
    )
    sp.set_defaults(func=mode_dt)


def parse_el(spa: _SubParsersAction) -> None:
    """Add the el mode subparser.

    :param spa: The subparser action used to add a new subparser to
        the main parser.
    :return: None.
    :rtype: NoneType
    """
    sp = spa.add_parser(
        'el',
        aliases=['escapelist', 'esclist',],
        description='List the registered escape schemes.'
    )
    sp.add_argument(
        '-d', '--description',
        help='Show the description for the character sets.',
        action='store_true'
    )
    sp.set_defaults(func=mode_el)


def parse_es(spa: _SubParsersAction) -> None:
    """Add the escape mode subparser.

    :param spa: The subparser action used to add a new subparser to
        the main parser.
    :return: None.
    :rtype: NoneType
    """
    valid_schemes = ', '.join(scheme for scheme in esc.get_schemes())

    sp = spa.add_parser(
        'es',
        aliases=['escape', 'esc',],
        description='Escape the string.'
    )
    sp.add_argument(
        'scheme',
        action='store',
        choices=esc.get_schemes(),
        default='url',
        help=(
            'The scheme to escape with. The valid schemes '
            f'are: {valid_schemes}.'
        ),
        metavar='scheme'
    )
    sp.add_argument(
        'base',
        help='The string to escape.',
        action='store',
        type=str
    )
    sp.set_defaults(func=mode_es)


def parse_fl(spa: _SubParsersAction) -> None:
    """Add the fl mode subparser.

    :param spa: The subparser action used to add a new subparser to
        the main parser.
    :return: None.
    :rtype: NoneType
    """
    sp = spa.add_parser(
        'fl',
        aliases=['formlist', 'flist',],
        description='List the registered normalization forms.'
    )
    sp.add_argument(
        '-d', '--description',
        help='Show the description for the character sets.',
        action='store_true'
    )
    sp.set_defaults(func=mode_fl)


def parse_nl(spa: _SubParsersAction) -> None:
    """Add the nl mode subparser.

    :param spa: The subparser action used to add a new subparser to
        the main parser.
    :return: None.
    :rtype: NoneType
    """
    valid_forms = ', '.join(form for form in nl.get_forms())

    sp = spa.add_parser(
        'nl',
        aliases=['normal',],
        description='Normalize a string.'
    )
    sp.add_argument(
        'form',
        choices=nl.get_forms(),
        help=(
            'The normalization form for the normalization. Valid '
            f'options are: {valid_forms}.'
        ),
        metavar='form'
    )
    sp.add_argument(
        'base',
        help='The base normalized string.',
        action='store',
        type=str
    )
    sp.set_defaults(func=mode_nl)


def parse_sh(spa: _SubParsersAction) -> None:
    """Add the shell mode subparser.

    :param spa: The subparser action used to add a new subparser to
        the main parser.
    :return: None.
    :rtype: NoneType
    """
    sp = spa.add_parser(
        'sh',
        aliases=['shell',],
        description='Run charex in an interactive shell.'
    )
    sp.set_defaults(func=mode_sh)


def invoke(
    cmd: str | None = None,
    p: ArgumentParser | None = None
) -> None:
    """Parse the arguments used to invoke the script and execute
    the script.
    """
    if not p:
        p = build_parser()
    if cmd:
        argv = split(cmd)
        args = p.parse_args(argv)
    else:
        args = p.parse_args()
    args.func(args)


# The interactive shell.
class Shell(Cmd):
    """A command shell for :mod:`charex`."""
    intro = (
        'Welcome to the charex shell.\n'
        'Press ? for a list of comands.\n'
    )
    prompt = 'charex> '

    def __init__(self, *args, **kwargs) -> None:
        self.parser = build_parser()
        super().__init__(*args, **kwargs)

    # Commands.
    def do_cd(self, arg):
        """Decode the given address in all codecs."""
        cmd = f'cd {arg}'
        self._run_cmd(cmd)

    def do_ce(self, arg):
        """Encode the given character in all codecs."""
        cmd = f'ce {arg}'
        self._run_cmd(cmd)

    def do_cl(self, arg):
        """List the registered character sets."""
        cmd = f'cl {arg}'
        self._run_cmd(cmd)

    def do_ct(self, arg):
        """Count denormalization results."""
        cmd = f'ct {arg}'
        self._run_cmd(cmd)

    def do_dm(self, arg):
        """Build a denormalization map."""
        cmd = f'dm {arg}'
        self._run_cmd(cmd)

    def do_dn(self, arg):
        """Denormalize the given string."""
        cmd = f'dn {arg}'
        self._run_cmd(cmd)

    def do_dt(self, arg):
        """Get details for the given character."""
        cmd = f'dt {arg}'
        self._run_cmd(cmd)

    def do_el(self, arg):
        """List the registered escape schemes."""
        cmd = f'el {arg}'
        self._run_cmd(cmd)

    def do_EOF(self, arg):
        """Exit the charex shell."""
        print()
        print('Exiting charex.')
        print()
        return True

    def do_es(self, arg):
        """Escape the string."""
        cmd = f'es {arg}'
        self._run_cmd(cmd)

    def do_fl(self, arg):
        """List the registered normalization forms."""
        cmd = f'fl {arg}'
        self._run_cmd(cmd)

    def do_help(self, arg):
        """Display command list."""
        if not arg:
            print('The following commands are available:')
            print()
            cmds = (
                cmd for cmd in dir(self)
                if cmd.startswith('do')
                and not cmd.endswith('EOF')
                and not cmd.endswith('eader')
            )
            for cmd in cmds:
                meth = getattr(self, cmd)
                print(f'*  {cmd[3:]}: {meth.__doc__}')
            print()
            print('For help on individual commands, use "help {command}".')
            print()

        else:
            super().do_help(arg)

    def do_nl(self, arg):
        """Normalize the given string."""
        cmd = f'nl {arg}'
        self._run_cmd(cmd)

    def do_xt(self, arg):
        """Exit the charex shell."""
        print('Exiting charex.')
        print()
        return True

    # Command help.
    def help_cd(self):
        """Help for the cd command."""
        cmd = f'cd -h'
        self._run_cmd(cmd)

    def help_ce(self):
        cmd = f'ce -h'
        self._run_cmd(cmd)

    def help_cl(self):
        cmd = f'cl -h'
        self._run_cmd(cmd)

    def help_ct(self):
        """Help for the ct command."""
        cmd = f'ct -h'
        self._run_cmd(cmd)

    def help_dn(self):
        """Help for the dn command."""
        cmd = f'dn -h'
        self._run_cmd(cmd)

    def help_dm(self):
        """Help for the dm command."""
        cmd = f'dm -h'
        self._run_cmd(cmd)

    def help_dt(self):
        """Help for the dt command."""
        cmd = f'dt -h'
        self._run_cmd(cmd)

    def help_el(self):
        """Help for the el command."""
        cmd = f'el -h'
        self._run_cmd(cmd)

    def help_es(self):
        """Help for the es command."""
        cmd = f'es -h'
        self._run_cmd(cmd)

    def help_fl(self):
        """Help for the fl command."""
        cmd = f'fl -h'
        self._run_cmd(cmd)

    def help_nl(self):
        """Help for the nl command."""
        cmd = f'nl -h'
        self._run_cmd(cmd)

    def help_xt(self):
        lines = util.read_resource('help_xt')
        print(''.join(lines))

    # Private methods.
    def _run_cmd(self, cmd):
        """Run the given command."""
        try:
            invoke(cmd, self.parser)
        except SystemExit as ex:
            print()


# Mode registry.
modes = {
    name: globals()[name]
    for name in dir()
    if name.startswith('mode_')
}


def describe_modes() -> str:
    """Return a description of the operating modes."""
    global modes
    width, height = get_terminal_size((80, 20))
    text = wrap((
        'The following are brief desciptions of each of the '
        'available options for the mode:'
    ), width)
    result = '\n'.join(text) + '\n\n'
    for mode in modes:
        name = mode.split('_')[1]
        doc = modes[mode].__doc__
        descr = doc.split('\n\n')[0]
        result += f'  * {name}: {descr}\n'
    return result


def list_modes() -> str:
    """Return a description of the available modes."""
    global modes
    result = 'The following modes are available: '
    names = ', '.join(mode.split('_')[1] for mode in modes)
    result += names
    return result



"""
denormal
~~~~~~~~

Functions for reversing normalization of string.
"""
from collections.abc import Sequence
from math import prod
from random import choice, seed

from charex.charex import Character


# Functions.
def count_denormalizations(
    base: str,
    form: str,
    maxdepth: int | None = None
) -> int:
    """Determine the number of denormalizations that exist for the string.

    :param base: The :class:`str` to denormalize.
    :param form: The Unicode normalization form to denormalize from.
        Valid values are: casefold, nfc, nfd, nfkc, nfkd.
    :param maxdepth: (Optional.) How many individual characters to use
        when denormalizing the base. This is used to limit the total
        number of denormalizations of the overall base.
    :return: The number of denormalizations as an :class:`int`.
    :rtype: int
    """
    chars = (Character(c) for c in base)
    counts = []
    for char in chars:
        count = len(char.denormalize(form))
        if count == 0:
            count = 1
        if maxdepth and count > maxdepth:
            count = maxdepth
        counts.append(count)
    return int(prod(counts))


def denormalize(
    base: str,
    form: str,
    maxdepth: int | None = None,
    maxresults: int | None = None,
    random: bool = False,
    seed_: bytes | int | str = ''
) -> tuple[str, ...]:
    """Denormalize a string.

    :param base: The :class:`str` to denormalize.
    :param form: The Unicode normalization form to denormalize from.
        Valid values are: casefold, nfc, nfd, nfkc, nfkd.
    :param maxdepth: (Optional.) How many individual characters to use
        when denormalizing the base. This is used to limit the total
        number of denormalizations of the overall base.
    :param maxresults: (Optional.) The maximum number of results to
        return. Default behavior varies based on the `random` parameter.
        If `random` is `False`, default is to return all possible
        denormalizattions. Otherwise, the default is to return one.
    :param random: (Optional.) Whether to pick randomly from the
        possible denormalization results. Defaults to false.
    :param seed: (Optional.) A seed value for the random number generator.
        Defaults to not seeding the generator.
    :return: The denormalizations as a :class:`tuple`.
    :rtype: tuple
    """
    if random:
        return random_denormalize(base, form, maxresults, seed_)

    # Get the denormalized forms of the first character.
    char = Character(base[0])
    dechars = list(char.denormalize(form))

    # If there are no denormalized forms, then it is the denormalized form.
    if not dechars:
        dechars = [char.value,]

    # Limit the number of permutations by limiting the number of
    # denormalized forms we are looking at.
    if maxdepth and len(dechars) > maxdepth:
        dechars = dechars[:maxdepth]

    # If there are more characters left, use recursion to get the
    # permutations for those characters, then get the permutations
    # with this character.
    if base[1:]:
        results = []
        tails = denormalize(base[1:], form, maxdepth, maxresults)
        for dechar in dechars:
            for tail in tails:
                results.append(dechar + tail)

    # If there are no characters left, then the permutations are just
    # the denormalized forms of the character.
    else:
        results = dechars

    # Truncate to the maximum results and return.
    if maxresults:
        results = results[:maxresults]
    return tuple(results)


def random_denormalize(
    base: str,
    form: str,
    maxresults: int | None = None,
    seed_: bytes | int | str = ''
) -> tuple[str, ...]:
    """Randomly denormalize a string.

    :param base: The :class:`str` to denormalize.
    :param form: The Unicode normalization for to denormalize from.
        Valid values are: NFC, NFD, NFKC, NFKD.
    :param maxresults: (Optional.) The maximum number of results to
        return. Default behavior varies based on the `random` parameter.
        If `random` is `False`, default is to return all possible
        denormalizattions. Otherwise, the default is to return one.
    :param seed: (Optional.) A seed value for the random number generator.
        Defaults to not seeding the generator.
    :return: The denormalizations as a :class:`tuple`.
    :rtype: tuple
    """
    # Ensure at least one result is returned.
    if not maxresults:
        maxresults = 1

    # Seeding the RNG allows for repeatability in testing.
    if seed_:
        seed(seed_)

    # Get the denormalized forms for all the characters in the string.
    chars = [
        denormalize(char, form)
        for char in base
    ]

    # Randomly pick from the possible denormalizations for each character
    # when creating the denormalized strings, then return the results.
    results = []
    for _ in range(maxresults):
        result = ''.join(choice(char) for char in chars)
        results.append(result)
    return tuple(results)



"""
test_main
~~~~~~~~~

Unit tests for the mainline of the `charex` package.
"""
import sys

from charex import __main__ as m
from charex import escape as esc
from charex import normal as nl


# Test cd mode.
def test_cd(capsys):
    """Called with an hex string, charset mode should return the character
    or characters that hex string becomes in each of the known character
    sets.
    """
    with open('tests/data/charset_mode_41.txt') as fh:
        exp = fh.read()
    cmd = (
        'python -m charex',
        'cd',
        '0x41'
    )
    cli_test(exp, cmd, capsys)


def test_cd_binary(capsys):
    """Called with a number prefixed with "0b", cd mode should interpret
    the base string as binary and return the character or characters that
    binary string becomes in each of the known character sets.
    """
    with open('tests/data/charset_mode_41.txt') as fh:
        exp = fh.read()
    cmd = (
        'python -m charex',
        'cd',
        '0b01000001',
    )
    cli_test(exp, cmd, capsys)


def test_cd_code_point(capsys):
    """Called with a character, cd mode should interpret the base
    string as binary and return the character or characters that hex
    string becomes in each of the known character sets.
    """
    with open('tests/data/charset_mode_41.txt') as fh:
        exp = fh.read()
    cmd = (
        'python -m charex',
        'cd',
        'A',
    )
    cli_test(exp, cmd, capsys)


def test_cd_control_character(capsys):
    """Called with an hex string, cd mode should return the character
    or characters that hex string becomes in each of the known character
    sets. If the hex string becomes a control character, print the symbol
    for that character rather than the character itself.
    """
    with open('tests/data/charset_mode_0a.txt') as fh:
        exp = fh.read()
    cmd = (
        'python -m charex',
        'cd',
        '0x0a'
    )
    cli_test(exp, cmd, capsys)


def test_cd_no_character(capsys):
    """Called with an hex string, cd mode should return the character
    or characters that hex string becomes in each of the known character
    sets. If some character sets do not have characters for the given
    address, that should be indicated in the output.
    """
    with open('tests/data/charset_mode_e9.txt') as fh:
        exp = fh.read()
    cmd = (
        'python -m charex',
        'cd',
        '0xe9'
    )
    cli_test(exp, cmd, capsys)


# Test ce mode.
def test_ce(capsys):
    """Called with an character, ce mode should return the address
    for the character in each of the registered character sets.
    """
    with open('tests/data/charset_mode_r.txt') as fh:
        exp = fh.read()
    cmd = (
        'python -m charex',
        'ce',
        'A',
    )
    cli_test(exp, cmd, capsys)


# Test cl mode.
def test_cl(capsys):
    """When invoked, cl mode should return a list of registered
    character set codecs.
    """
    with open('tests/data/charsetlist.txt') as fh:
        exp = fh.read()
    cmd = (
        'python -m charex',
        'cl'
    )
    cli_test(exp, cmd, capsys)


def test_cl_description(capsys):
    """When invoked with -d, cl mode should return a list of
    registered character set codecs and a brief description of each
    one.
    """
    with open('tests/data/charsetlist_d.txt') as fh:
        exp = fh.read()
    cmd = (
        'python -m charex',
        'cl',
        '-d',
    )
    cli_test(exp, cmd, capsys)


# Test ct mode.
def test_ct(capsys):
    """Invoked with a normalization form and a base string, ct mode
    should print the number of denormalizations using the given form to
    stdout.
    """
    exp = '120,270,240\n\n'
    cmd = (
        'python -m charex',
        'ct',
        'nfkd',
        '<script>'
    )
    cli_test(exp, cmd, capsys)


def test_ct_maxdepth(capsys):
    """Invoked with "-m" and an integer, ct mode limit the number of
    denormalizations per character to the given integer and print the
    number of denormalizations using the given form to stdout.
    """
    exp = '256\n\n'
    cmd = (
        'python -m charex',
        'ct',
        'nfkd',
        '<script>',
        '-m', '2',
    )
    cli_test(exp, cmd, capsys)


# Test dn mode.
def test_dn(capsys):
    """Invoked with a normalization form and a base string, dn mode
    should print the denormalizations for the base string to stdout.
    """
    # Expected result.
    exp = (
        '\ufe64\ufe63\ufe65\n'
        '\ufe64\ufe63\uff1e\n'
        '\ufe64\uff0d\ufe65\n'
        '\ufe64\uff0d\uff1e\n'
        '\uff1c\ufe63\ufe65\n'
        '\uff1c\ufe63\uff1e\n'
        '\uff1c\uff0d\ufe65\n'
        '\uff1c\uff0d\uff1e\n'
        '\n'
    )
    cmd = (
        'python -m charex',
        'dn',
        'nfkd',
        '<->'
    )
    cli_test(exp, cmd, capsys)


def test_dn_number(capsys):
    """Invoked with -n and an integer, dn mode should return no
    more than that number of results.
    """
    exp = (
        '\ufe64\ufe63\ufe65\n'
        '\ufe64\ufe63\uff1e\n'
        '\ufe64\uff0d\ufe65\n'
        '\ufe64\uff0d\uff1e\n'
        '\n'
    )
    cmd = (
        'python -m charex',
        'dn',
        'nfkd',
        '<->',
        '-n', '4'
    )
    cli_test(exp, cmd, capsys)


def test_dn_random(capsys):
    """Called with -r, dn mode should return a randomly
    denormalize the string.
    """
    exp = (
        '﹤－﹥\n'
        '\n'
    )
    cmd = (
        'python -m charex',
        'dn',
        'nfkd',
        '<->',
        '-r',
        '-s', 'spam'
    )
    cli_test(exp, cmd, capsys)


# Test dt mode.
def test_dt(capsys):
    """Invoked with a character, details mode should print the details
    for the character.
    """
    with open('tests/data/details_mode_A.txt') as fh:
        exp = fh.read()
    cmd = (
        'python -m charex',
        'dt',
        'A'
    )
    cli_test(exp, cmd, capsys)


# Test el mode.
def test_el(capsys):
    """When invoked, el mode returns a list of the registered
    escape schemes.
    """
    exp = '\n'.join(scheme for scheme in esc.schemes) + '\n\n'
    cmd = (
        'python -m charex',
        'el',
    )
    cli_test(exp, cmd, capsys)


def test_el_description(capsys):
    """When invoked with -d, cl mode should return a list of
    registered escape schemes and a brief description of each
    one.
    """
    with open('tests/data/el_d.txt') as fh:
        exp = fh.read()
    cmd = (
        'python -m charex',
        'el',
        '-d',
    )
    cli_test(exp, cmd, capsys)


# Test es mode.
def test_es(capsys):
    """Invoked with a scheme and a base string, escape mode should
    escape the string using the given scheme and print the escaped
    string.
    """
    exp = '%41\n\n'
    cmd = (
        'python -m charex',
        'es',
        'url',
        'A',
    )
    cli_test(exp, cmd, capsys)


# Test fl mode.
def test_fl(capsys):
    """When invoked, fl mode returns a list of the registered
    normalization forms.
    """
    exp = '\n'.join(form for form in nl.forms) + '\n\n'
    cmd = (
        'python -m charex',
        'fl',
    )
    cli_test(exp, cmd, capsys)


def test_fl_description(capsys):
    """When invoked with -d, fl mode should return a list of
    registered normalization forms and a brief description of each
    one.
    """
    with open('tests/data/fl_d.txt') as fh:
        exp = fh.read()
    cmd = (
        'python -m charex',
        'fl',
        '-d',
    )
    cli_test(exp, cmd, capsys)


# Test nl mode.
def test_nl(capsys):
    """When invoked with a normalization form and a base string,
    nl mode returns the normalization of the base string using the
    given form.
    """
    exp = 'A\n\n'
    cmd = (
        'python -m charex',
        'nl',
        'nfkc',
        '\u24b6',
    )
    cli_test(exp, cmd, capsys)


# Utility functions.
def cli_test(exp, cmd, capsys):
    """Test command line invocation."""
    # Test set up.
    orig_cmd = sys.argv
    sys.argv = cmd

    # Run test.
    m.sh.invoke()

    # Gather actual result and compare.
    captured = capsys.readouterr()
    assert captured.out == exp

    # Test tear down.
    sys.argv = orig_cmd



"""
escape
~~~~~~

Character escape schemes.
"""
from collections.abc import Callable
from json import loads

from charex import util


# Registry.
schemes: dict[str, Callable[[str, str], str]] = {}


# Caches.
cached_entities: dict[str, str] = {}


# Registration.
class reg_escape:
    """A decorator for registering escape schemes.

    :param key: The name the escape sequence is registered under.
    """
    def __init__(self, key: str) -> None:
        self.key = key

    def __call__(
        self,
        fn: Callable[[str, str], str]
    ) -> Callable[[str, str], str]:
        schemes[self.key] = fn
        return fn


# Exceptions.
class EscapeError(ValueError):
    """The escape scheme could not escape the character."""


# Utility functions.
def get_named_entity(char: str) -> str:
    """Get a named entity from the HTML entity data."""
    lines = util.read_resource('entities')
    json = ''.join(lines)
    data = loads(json)
    by_char = {data[key]['characters']: key for key in data}
    try:
        cached_entities[char] = by_char[char]
    except KeyError:
        cached_entities[char] = escape_htmldec(char, '')
    return cached_entities[char]


def get_description(schemekey: str) -> str:
    """Get the description for the scheme.

    :param schemekey: The key for the scheme in the scheme registry.
    :return: The description as a :class:`str`.
    :rtype: str
    """
    scheme = schemes[schemekey]
    return util.get_description_from_docstring(scheme)


def get_schemes() -> tuple[str, ...]:
    """Return the keys of the registered escape schemes.

    :return: The scheme keys as a :class:`tuple`.
    :rtype: tuple
    """
    return tuple(scheme for scheme in schemes)


def hex_byte_escape(char: str) -> str:
    """Perform the common single hexadecimal byte escape on the
    character.

    :param char: The character to escape.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    n = ord(char)
    if n > 0xFF:
        raise EscapeError('Cannot escape characters over 0xFF.')
    return f'\\x{n:02x}'


def lookup_escape(char: str, table: dict[str, str]) -> str:
    """Perform a table lookup to escape the character.

    :param char: The character to escape.
    :param table: The table for the lookup.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    try:
        return table[char]
    except KeyError:
        raise EscapeError('Character not in table.')


def octal_escape(char: str) -> str:
    """Perform the common octal escape on the character.

    :param char: The character to escape.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    n = ord(char)
    if n > 0o377:
        raise EscapeError('Cannot escape characters over 0o377.')
    return f'\\{n:o}'


def unicode_2_byte_escape(char: str) -> str:
    """Perform the common Unicode two byte escape on the
    character.

    :param char: The character to escape.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    n = ord(char)
    if n > 0xFFFF:
        raise EscapeError('Cannot escape characters over 0xFFFF.')
    return f'\\u{n:04x}'


def unicode_utf16_escape(char: str) -> str:
    """Perform the common Unicode UTF-16 escape on the character.

    :param char: The character to escape.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    try:
        return unicode_2_byte_escape(char)
    except EscapeError:
        b = char.encode('utf_16_be')
        result = ''
        for i in range(0, len(b), 2):
            result += '\\u'
            result += f'{b[i]:02x}'
            result += f'{b[i + 1]:02x}'
        return result


# Escape schemes.
@reg_escape('c')
def escape_c(char: str, codec: str) -> str:
    """Escape scheme for C escape sequences as defined by C17.

    This is derived from the Wikipedia list, since I don't have access
    to the C17 specification.

    :param char: The character to escape.
    :param codec: Unused.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    table = {
        '\u0007': r'\a',
        '\u0008': r'\b',
        '\u000c': r'\f',
        '\u000a': r'\n',
        '\u000d': r'\r',
        '\u0009': r'\t',
        '\u000b': r'\v',
        # '\u001b': r'\e',    # Non-standard, supported by gcc, clang, tcc.
        '\u0027': r"\'",
        '\u0022': r'\"',
        '\u003f': r'\?',
        '\u005c': r'\\',
    }
    try:
        return lookup_escape(char, table)
    except EscapeError:
        return escape_co(char, codec)


@reg_escape('co')
def escape_co(char: str, codec: str) -> str:
    """Escape scheme for C octal escape sequences as defined by C17.

    This is derived from the Wikipedia list, since I don't have access
    to the C17 specification.

    :param char: The character to escape.
    :param codec: Unused.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    try:
        return octal_escape(char)
    except EscapeError:
        return escape_cu(char, codec)


@reg_escape('cu')
def escape_cu(char: str, codec: str) -> str:
    """Escape scheme for C Unicode escape sequences as defined by C17.

    :param char: The character to escape.
    :param codec: Unused.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    try:
        return unicode_2_byte_escape(char)
    except EscapeError:
        return escape_culong(char, codec)


@reg_escape('culong')
def escape_culong(char: str, codec: str) -> str:
    """Escape scheme for four byte C Unicode escape sequences as
    defined by C17.

    :param char: The character to escape.
    :param codec: Unused.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    x = ord(char)
    return f'\\U{x:08x}'


@reg_escape('html')
def escape_html(char: str, codec: str) -> str:
    """Escape scheme for HTML named character references. It will return
    the decimal numeric character references if no named entity exists.

    :param char: The character to escape.
    :param codec: Unused.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    try:
        return cached_entities[char]
    except KeyError:
        return get_named_entity(char)


@reg_escape('htmldec')
def escape_htmldec(char: str, codec: str) -> str:
    """Escape scheme for HTML decimal numeric character references.

    :param char: The character to escape.
    :param codec: Unused.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    n = ord(char)
    return f'&#{n};'


@reg_escape('htmlhex')
def escape_htmlhex(char: str, codec: str) -> str:
    """Escape scheme for HTML hexadecimal numeric character references.

    :param char: The character to escape.
    :param codec: Unused.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    n = ord(char)
    return f'&#x{n:x};'


@reg_escape('java')
def escape_java(char: str, codec: str) -> str:
    """Escape scheme for Java encoding, based on the Java SE
    Specification.

    The specification can be found `here.`_

    .. _here: https://docs.oracle.com/javase/specs/jls/se20/html/jls-3.html

    :param char: The character to escape.
    :param codec: The character set to use when encoding the character.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    table = {
        '\u0008': r'\b',
        '\u0020': r'\s',
        '\u000c': r'\f',
        '\u000a': r'\n',
        '\u000d': r'\r',
        '\u0009': r'\t',
        '\u0027': r"\'",
        '\u0022': r'\"',
        '\u005c': r'\\',
    }
    try:
        return lookup_escape(char, table)
    except EscapeError:
        return escape_javao(char, codec)


@reg_escape('javao')
def escape_javao(char: str, codec: str) -> str:
    """Escape scheme for Java octal encoding, based on the Java SE
    Specification.

    The specification can be found `here.`_

    .. _here: https://docs.oracle.com/javase/specs/jls/se20/html/jls-3.html

    :param char: The character to escape.
    :param codec: The character set to use when encoding the character.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    try:
        return octal_escape(char)
    except EscapeError:
        return escape_javau(char, codec)


@reg_escape('javau')
def escape_javau(char: str, codec: str) -> str:
    """Escape scheme for Java Unicode encoding, based on the Java SE
    Specification.

    The specification can be found `here.`_

    .. _here: https://docs.oracle.com/javase/specs/jls/se20/html/jls-3.html

    :param char: The character to escape.
    :param codec: The character set to use when encoding the character.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    return unicode_utf16_escape(char)


@reg_escape('js')
def escape_js(char: str, codec: str) -> str:
    """Escape scheme for JavaScript encoding, based on the ECMA-262
    Specification.

    The specification can be found `here.`_

    .. _here: https://262.ecma-international.org/13.0/\
#sec-literals-string-literals

    :param char: The character to escape.
    :param codec: The character set to use when encoding the character.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    table = {
        '\u0008': r'\b',
        '\u0009': r'\t',
        '\u000a': r'\n',
        '\u000b': r'\v',
        '\u000c': r'\f',
        '\u000d': r'\r',
        '\u0022': r'\"',
        '\u005c': r'\\',
    }
    try:
        return lookup_escape(char, table)
    except EscapeError:
        return escape_jso(char, codec)


@reg_escape('jso')
def escape_jso(char: str, codec: str) -> str:
    """Escape scheme for JavaScript octal encoding, based on the
    ECMA-262 Specification.

    The specification can be found `here.`_

    .. _here: https://262.ecma-international.org/13.0/\
#sec-literals-string-literals

    :param char: The character to escape.
    :param codec: The character set to use when encoding the character.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    try:
        return octal_escape(char)
    except EscapeError:
        return escape_jsu(char, codec)


@reg_escape('jsu')
def escape_jsu(char: str, codec: str) -> str:
    """Escape scheme for JavaScript unicode encoding, based on the
    ECMA-262 Specification.

    The specification can be found `here.`_

    .. _here: https://262.ecma-international.org/13.0/\
#sec-literals-string-literals

    :param char: The character to escape.
    :param codec: Unused.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    try:
        return unicode_2_byte_escape(char)
    except EscapeError:
        return escape_jscp(char, codec)


@reg_escape('jscp')
def escape_jscp(char: str, codec: str) -> str:
    """Escape scheme for JavaScript code point encoding, based on the
    ECMA-262 Specification.

    The specification can be found `here.`_

    .. _here: https://262.ecma-international.org/13.0/\
#sec-literals-string-literals

    :param char: The character to escape.
    :param codec: The character set to use when encoding the character.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    x = ord(char)
    return f'\\u{{{x:x}}}'


@reg_escape('json')
def escape_json(char: str, codec: str) -> str:
    """Escape scheme for JSON encoding, based on the ECMA-404
    Specification.

    The specification can be found `here.`_

    .. _here: https://www.ecma-international.org/publications-and-standards/\
standards/ecma-404/ECMA-404_2nd_edition_december_2017.pdf

    :param char: The character to escape.
    :param codec: The character set to use when encoding the character.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    table = {
        '\u0022': r'\"',
        '\u005c': r'\\',
        '\u002f': r'\/',
        '\u0008': r'\b',
        '\u000c': r'\f',
        '\u000a': r'\n',
        '\u000d': r'\r',
        '\u0009': r'\t',
    }
    if char in table:
        return table[char]
    return escape_jsonu(char, codec)


@reg_escape('jsonu')
def escape_jsonu(char: str, codec: str) -> str:
    """Escape scheme for JSON Unicode encoding, based on the ECMA-404
    Specification.

    The specification can be found `here.`_

    .. _here: https://www.ecma-international.org/publications-and-standards/\
standards/ecma-404/ECMA-404_2nd_edition_december_2017.pdf

    :param char: The character to escape.
    :param codec: The character set to use when encoding the character.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    return unicode_utf16_escape(char)


@reg_escape('sql')
def escape_sql(char: str, codec: str) -> str:
    """Escape scheme for MySQL encoding, based on the MySQL
    Specification.

    The specification can be found `here.`_

    .. _here: https://dev.mysql.com/doc/refman/8.0/en/string-literals.html

    :param char: The character to escape.
    :param codec: The character set to use when encoding the character.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    table = {
        '\u0000': r'\0',
        '\u0027': r"\'",
        '\u0022': r'\"',
        '\u0008': r'\b',
        '\u000a': r'\n',
        '\u000d': r'\r',
        '\u0009': r'\t',
        '\u0026': r'\Z',
        '\u005c': r'\\',
        '\u0025': r'\%',
        '\u005f': r'\_',
    }
    try:
        return lookup_escape(char, table)
    except EscapeError:
        return char


@reg_escape('sqldq')
def escape_sqldq(char: str, codec: str) -> str:
    """Escape scheme for MySQL encoding, based on the MySQL
    Specification. This escapes qoutes by doubling them rather
    than using a backslash.

    The specification can be found `here.`_

    .. _here: https://dev.mysql.com/doc/refman/8.0/en/string-literals.html

    :param char: The character to escape.
    :param codec: The character set to use when encoding the character.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    table = {
        '\u0000': r'\0',
        '\u0027': r"''",
        '\u0022': r'""',
        '\u0008': r'\b',
        '\u000a': r'\n',
        '\u000d': r'\r',
        '\u0009': r'\t',
        '\u0026': r'\Z',
        '\u005c': r'\\',
        '\u0025': r'\%',
        '\u005f': r'\_',
    }
    try:
        return lookup_escape(char, table)
    except EscapeError:
        return char


@reg_escape('url')
def escape_url(char: str, codec: str) -> str:
    """Escape scheme for URL percent encoding.

    :param char: The character to escape.
    :param codec: The character set to use when encoding the character.
    :return: The escaped character as a :class:`str`.
    :rtype: str
    """
    b = char.encode(codec)
    octets = [f'%{x:02x}'.upper() for x in b]
    return ''.join(x for x in octets)


# Bulk escape.
def escape(s: str, schemekey: str, codec: str = 'utf8') -> str:
    """Escape the string with the scheme.

    :param s: The string to escape.
    :param scheme: The key in the `schemes` :class:`dict` to use for
        the escaping.
    :param codec: The character set codec to use when escaping the
        characters.
    :return: The escaped :class:`str`.
    :rtype: str
    """
    scheme = schemes[schemekey]
    return ''.join(scheme(char, codec) for char in s)



"""
test_escape
~~~~~~~~~~~

Unit tests for :mod:`charex.escape`.
"""
from charex import escape as esc


# Test escape.
def test_escape():
    """Given a string and the key for an escape scheme,
    :func:`charex.escape.escape` should return a string with every
    character escaped using the given string.
    """
    exp = '%73%70%61%6D'
    assert esc.escape('spam', 'url')


# Test escape_c.
def test_escape_c():
    """Given a character and a codec, return the C/C++ escape
    sequence for the character. Note: the codec doesn't do anything,
    it's just here for compatibility.
    """
    exp = r'\n'
    assert esc.escape_c('\u000a', '') == exp


# Test escape_co.
def test_escape_co():
    """Given a character and a codec, return the C/C++ octal escape
    sequence for the character. Note: the codec doesn't do anything,
    it's just here for compatibility.
    """
    exp = r'\141'
    assert esc.escape_co('a', '') == exp


# Test escape_cu.
def test_escape_cu():
    """Given a character and a codec, return the C/C++ Unicode escape
    sequence for the character. Note: the codec doesn't do anything,
    it's just here for compatibility.
    """
    exp = r'\u0061'
    assert esc.escape_cu('a', '') == exp


def test_escape_cu_4_byte():
    """Given a character and a codec, return the C/C++ Unicode escape
    sequence for the character. If the character is over two bytes long,
    the sequence should use a capital U and eight digits. Note: the
    codec doesn't do anything, it's just here for compatibility.
    """
    exp = r'\U00010000'
    assert esc.escape_cu('\U00010000', '') == exp


# Test escape_culong.
def test_escape_culong():
    """Given a character and a codec, return the C/C++ Unicode escape
    sequence for the character. Note: the codec doesn't do anything,
    it's just here for compatibility.
    """
    exp = r'\U00000061'
    assert esc.escape_culong('a', '') == exp


# Test escape_html.
def test_html():
    """Given a character and a codec, return the HTML named entity for
    the given character. Note: codec doesn't do anything, it's just
    included for compatibility."""
    exp = "&qout;"
    assert esc.escape_html('"', '')


def test_html_no_name():
    """Given a character and a codec, return the HTML named entity for
    the given character. Note: codec doesn't do anything, it's just
    included for compatibility. If there is no specific named entity,
    return the decimal encoded entity."""
    exp = "&#97;"
    assert esc.escape_html('a', '')


# Test escape_htmldec.
def test_escape_htmldec():
    """Given a character and a codec, return the HTML entity for the
    address of that character.
    """
    exp = '&#97;'
    assert esc.escape_htmldec('a', 'utf8')


# Test escape_htmlhex.
def test_escape_htmlhex():
    """Given a character and a codec, return the HTML hexadecimal
    entity for the given character. Note: codec doesn't do anything,
    it's just included for compatibility.
    """
    exp = '&#x61;'
    assert esc.escape_htmlhex('a', '')


# Test escape_java.
def test_escape_java():
    """Given a character and a codec, return the Java escape
    sequence for the character. Note: the codec doesn't do anything,
    it's just here for compatibility.
    """
    exp = r'\n'
    assert esc.escape_java('\u000A', '') == exp


# Test escape_javao.
def test_escape_javao():
    """Given a character and a codec, return the Java octal escape
    sequence for the character. Note: the codec doesn't do anything,
    it's just here for compatibility.
    """
    exp = r'\141'
    assert esc.escape_javao('a', '') == exp


# Test escape_javau.
def test_escape_javau():
    """Given a character and a codec, return the Java Unicode escape
    sequence for the character. Note: the codec doesn't do anything,
    it's just here for compatibility.
    """
    exp = r'\u0041'
    assert esc.escape_javau('A', '') == exp


def test_escape_javau_4_byte_character():
    """Given a character and a codec, return the Java Unicode escape
    sequence for the character. Note: the codec doesn't do anything,
    it's just here for compatibility. Characters over U+FFFF have to
    be returned as two characters because Java encodes strings with
    UTF-16.
    """
    exp = r'\ud800\udc00'
    assert esc.escape_javau('\U00010000', '') == exp


# Test escape_js.
def test_escape_js():
    """Given a character and a codec, return the JavaScript escape
    sequence for the character. Note: the codec doesn't do anything,
    it's just here for compatibility.
    """
    exp = r'\n'
    assert esc.escape_js('\u000A', '') == exp


# Test escape_jso.
def test_escape_jso():
    """Given a character and a codec, return the JavaScript octal escape
    sequence for the character. Note: the codec doesn't do anything,
    it's just here for compatibility.
    """
    exp = r'\141'
    assert esc.escape_jso('a', '') == exp


# Test escape_jsu.
def test_escape_jsu():
    """Given a character and a codec, return the JavaScript Unicode
    escape sequence for the character. Note: the codec doesn't do
    anything, it's just here for compatibility.
    """
    exp = r'\u0041'
    assert esc.escape_jsu('A', '') == exp


def test_escape_jsu_4_byte_character():
    """Given a character and a codec, return the JavaScript Unicode
    escape sequence for the character. Note: the codec doesn't do
    anything, it's just here for compatibility.
    """
    exp = r'\u{10000}'
    assert esc.escape_jsu('\U00010000', '') == exp


# Test escape_jscp.
def test_escape_jscp():
    """Given a character and a codec, return the JavaScript code point
    escape sequence for the character. Note: the codec doesn't do
    anything, it's just here for compatibility.
    """
    exp = r'\u{41}'
    assert esc.escape_jscp('A', '') == exp


# Test escape_json.
def test_escape_json():
    """Given a character and a codec, return the JSON escape
    sequence for the character. Note: the codec doesn't do anything,
    it's just here for compatibility.
    """
    exp = r'\n'
    assert esc.escape_json('\u000A', '') == exp


# Test escape_jsonu.
def test_escape_jsonu():
    """Given a character and a codec, return the JSON Unicode escape
    sequence for the character. Note: the codec doesn't do anything,
    it's just here for compatibility.
    """
    exp = r'\u0041'
    assert esc.escape_jsonu('A', '') == exp


def test_escape_jsonu_special():
    """Given a character and a codec, return the JSON Unicode escape
    sequence for the character. Note: the codec doesn't do anything,
    it's just here for compatibility.
    """
    exp = r'\u000a'
    assert esc.escape_jsonu('\n', '') == exp


def test_escape_jsonu_4_byte_character():
    """Given a character and a codec, return the Java Unicode escape
    sequence for the character. Note: the codec doesn't do anything,
    it's just here for compatibility. Characters over U+FFFF have to
    be returned as two characters because Java encodes strings with
    UTF-16.
    """
    exp = r'\ud800\udc00'
    assert esc.escape_jsonu('\U00010000', '') == exp


# Test escape_sql.
def test_escape_sql():
    """Given a character and a codec, return the SQL escape
    sequence for the character. Note: the codec doesn't do anything,
    it's just here for compatibility.
    """
    exp = r'\n'
    assert esc.escape_sql('\u000A', '') == exp


# Test escape_sqldq.
def test_escape_sqldq():
    """Given a character and a codec, return the SQL escape
    sequence for the character. Note: the codec doesn't do anything,
    it's just here for compatibility.
    """
    exp = r'""'
    assert esc.escape_sqldq('\u0022', '') == exp


# Test escape_url.
def test_escape_url():
    """Given a character and a codec, return the URL encoding for the
    address of that character.
    """
    exp = '%61'
    assert esc.escape_url('a', 'utf8')


# Tests for get_description.
def test_get_description():
    """Given the key for an escape scheme,
    :func:`charex.escape.get_description` should
    return the first paragraph of the docstring of
    that scheme.
    """
    # Expected value.
    exp = 'Eggs bacon.'

    # Test set up.
    @esc.reg_escape('__test_get_description')
    def escape_spam(char, codec):
        """Eggs bacon.

        Ham and toast.
        """
        return char

    # Run test and determine result.
    assert esc.get_description('__test_get_description') == exp

    # Test clean up.
    del esc.schemes['__test_get_description']


# Tests for get_schemes.
def test_get_schemes():
    """When called, :func:`charex.escape.get_schemes` should return
    the keys of the registered escape schemes as a :class:`tuple`.
    """
    exp = tuple(scheme for scheme in esc.schemes)
    assert esc.get_schemes() == exp



"""
test_charsets
~~~~~~~~~~~~~

Unit tests for :mod:`charex.charsets`.
"""
from charex import charsets as cs


# Tests for get_codecs.
def test_get_codecs():
    """WHen called, return a tuple of registered codecs."""
    exp = tuple(codec for codec in cs.codecs.keys())
    assert cs.get_codecs() == exp


# Test get_codec_description.
def test_get_codec_description():
    """Given the name of a character set codec, return the description
    of that codec.
    """
    codec = 'ascii'
    info = cs.codecs[codec]
    exp = info.description
    assert cs.get_codec_description(codec) == exp


# Tests for multiencode.
def test_multiencode():
    """Given a code point and a list of character sets, return the
    :class:`bytes` for that code point in each character set as a
    :class:`dict`.
    """
    exp = {
        'ascii': b'',
        'cp1252': b'\x93',
        'mac_roman': b'\xd2',
        'utf_8': b'\xe2\x80\x9c',
    }
    codecs = exp.keys()
    act = cs.multiencode('“', codecs)
    assert exp == act


def test_multiencode():
    """Given an integer and a sequence of strings that reference
    decoding codecs, :func:`charex.charsets.multiencode` returns
    the code point for each given codec as a :class:`dict`.
    """
    exp = {
        'ascii': '',
        'cp1252': 'é',
        'iso8859_7': 'ι',
        'utf_16_be': 'é',
        'utf_16_le': 'é',
        'utf_16': 'é',
    }
    codecs = exp.keys()
    act = cs.multidecode(0xe9, codecs)
    assert exp == act


# Tests for multidecode.
def test_multidecode_bytes():
    """Given bytes and a sequence of strings that reference
    decoding codecs, :func:`charex.charsets.multiencode` returns
    the code point for each given codec as a :class:`dict`.
    """
    exp = {
        'ascii': '',
        'cp1252': 'é',
        'iso8859_7': 'ι',
        'utf_16_be': 'é',
        'utf_16_le': 'é',
        'utf_16': 'é',
    }
    codecs = exp.keys()
    act = cs.multidecode(b'\xe9', codecs)
    assert exp == act


def test_multidecode_str():
    """Given a hex string and a sequence of strings that reference
    decoding codecs, :func:`charex.charsets.multiencode` returns
    the code point for each given codec as a :class:`dict`.
    """
    exp = {
        'ascii': '',
        'cp1252': 'é',
        'iso8859_7': 'ι',
        'utf_16_be': 'é',
        'utf_16_le': 'é',
        'utf_16': 'é',
    }
    codecs = exp.keys()
    act = cs.multidecode('e9', codecs)
    assert exp == act



"""
test_util
~~~~~~~~~

Unit test for :mod:`charex.util`.
"""
from charex import util


# Test cases.
# Tests for bin2bytes.
def test_bin2bytes():
    """Given a :class:`str` containing a representation of a binary
    number, return that number as :class:`bytes`.
    """
    exp = b'\xe9'
    value = '11101001'
    act = util.bin2bytes(value)
    assert act == exp


def test_bin2bytes_len_not_multiple_of_eight():
    """Given a :class:`str` containing a representation of a binary
    number, return that number as :class:`bytes`. If the length of the
    :class:`str` is not a multiple of eight, prepend enough zeros to
    make it a multiple of eight.
    """
    exp = b'\x29\xe9'
    value = '10100111101001'
    act = util.bin2bytes(value)
    assert act == exp


def test_bin2bytes_len_not_multiple_of_eight_little_endian():
    """Given a :class:`str` containing a representation of a binary
    number, return that number as :class:`bytes`. If the length of
    the :class:`str` is not a multiple of eight and the string is
    little endian, prepend enough zeros to the last byte to make it
    a multiple of eight.
    """
    exp = b'\xa7\x29'
    value = '10100111101001'
    act = util.bin2bytes(value, endian='little')
    assert act == exp


# Tests for hex2bytes.
def test_hex2bytes():
    """Given a :class:`str` containing a representation of a hexadecimal
    number, return that number as :class:`bytes`.
    """
    exp = b'\xbe\xef\xca\x5e'
    value = 'beefca5e'
    act = util.hex2bytes(value)
    assert act == exp


def test_hex2bytes_odd_length():
    """Given a :class:`str` containing a representation of a hexadecimal
    number, return that number as :class:`bytes`. If the :class:`str` has
    an odd number of characters, prepend a zero to the string before
    converting it to bytes.
    """
    exp = b'\x0e\xef\xca\x5e'
    value = 'eefca5e'
    act = util.hex2bytes(value)
    assert act == exp


def test_hex2bytes_odd_length_little_endian():
    """Given a :class:`str` containing a representation of a hexadecimal
    number, return that number as :class:`bytes`. If the :class:`str` has
    an odd number of characters and endian is "little" add a zero before
    the last character of the string.
    """
    exp = b'\xbe\xef\xca\x0e'
    value = 'beefcae'
    act = util.hex2bytes(value, endian='little')
    assert act == exp



"""
__main__
~~~~~~~~

Mainline for command line invocations of :mod:`charex`.
"""
from sys import argv

import charex.shell as sh


# Mainline.
def main() -> None:
    # If there were no arguments passed, drop into the command shell.
    if len(argv) < 2:
        sh.mode_sh(None)

    # Otherwise parse the arguments and execute.
    else:
        sh.invoke()


if __name__ == '__main__':
    main()



"""
__init__
~~~~~~~~

Initialization for the :mod:`charex` package.
"""
from charex.charex import Character
from charex.charsets import get_codecs, multidecode, multiencode
from charex.denormal import count_denormalizations, denormalize
from charex.escape import get_schemes
from charex.escape import escape as escape_text
from charex.normal import get_forms, normalize
